/*
**  NGPT - Next Generation POSIX Threading
**  Copyright (c) 2001 IBM Corporation <babt@us.ibm.com>
**  Portions Copyright (c) 1999-2000 Ralf S. Engelschall <rse@engelschall.com>
**
**  This file is part of NGPT, a non-preemptive thread scheduling
**  library which can be found at http://www.ibm.com/developer.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
**  Lesser General Public License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
**  USA.
**
**  pth_sched.c: Pth thread scheduler, the real heart of Pth
*/
                             /* ``Recursive, adj.;
                                  see Recursive.''
                                     -- Unknown   */
#include "pth_p.h"

intern struct pth_descr_st pth_native_list[PTH_MAX_NATIVE_THREADS];
				    /* complete list of native thread descriptors.	*/
intern pth_t        pth_main;       /* the main thread					*/
intern pth_t        pth_fork_thread;/* the fork thread (main thread of fork process)	*/
intern pth_pqueue_t pth_NQ;         /* queue of new threads				*/
intern pth_pqueue_t pth_RQ;         /* queue of threads ready to run			*/
intern pth_pqueue_t pth_WQ;         /* queue of threads waiting for mutex related event */
intern pth_pqueue_t pth_SQ;         /* queue of suspended threads			*/
intern pth_pqueue_t pth_DQ;         /* queue of terminated threads			*/

intern float        pth_loadval;    /* average scheduler load value			*/

static pth_time_t   pth_loadticknext;
static pth_time_t   pth_loadtickgap = PTH_TIME(1,0);

intern sigset_t	    pth_sigblock;
intern pth_qlock_t  pth_native_lock;
intern pth_qlock_t  pth_sig_lock;
intern pth_qlock_t  pth_scheduler_lock;

intern pth_threads_list_t   pth_threads_queue;
intern int		    pth_threads_count;

extern int sigandset(sigset_t *, __const sigset_t *, __const sigset_t *) __THROW;

intern void pth_allthread_mask(void)
{
    sigset_t tmp;
    pth_t t;
    int slot = 0;
    pid_t tid = current_tid();

    /* calculate global signal mask based on all threads sigmask */
    sigfillset(&tmp);

    _pth_acquire_lock(&(pth_NQ.q_lock), tid);
    for (t = pth_pqueue_head(&pth_NQ); t != NULL;
         t = pth_pqueue_walk(&pth_NQ, t, PTH_WALK_NEXT)) {
        sigandset(&tmp, &tmp, &t->mctx.sigs);
    }
    _pth_release_lock(&(pth_NQ.q_lock), tid);

    _pth_acquire_lock(&(pth_RQ.q_lock), tid);
    for (t = pth_pqueue_head(&pth_RQ); t != NULL;
         t = pth_pqueue_walk(&pth_RQ, t, PTH_WALK_NEXT)) {
        sigandset(&tmp, &tmp, &t->mctx.sigs);
    }
    _pth_release_lock(&(pth_RQ.q_lock), tid);

    _pth_acquire_lock(&(pth_WQ.q_lock), tid);
    for (t = pth_pqueue_head(&pth_WQ); t != NULL;
         t = pth_pqueue_walk(&pth_WQ, t, PTH_WALK_NEXT)) {
        sigandset(&tmp, &tmp, &t->mctx.sigs);
    }
    _pth_release_lock(&(pth_WQ.q_lock), tid);

    while (pth_native_list[slot].is_used) {
	pth_descr_t ds = &pth_native_list[slot++];
	_pth_acquire_lock(&(ds->wait_queue.q_lock), tid);
	for (t = pth_pqueue_head(&(ds->wait_queue)); t != NULL; 
	     t = pth_pqueue_walk(&(ds->wait_queue), t, PTH_WALK_NEXT))
	    sigandset(&tmp, &tmp, &t->mctx.sigs);
	_pth_release_lock(&(ds->wait_queue.q_lock), tid);
    }

    _pth_acquire_lock(&(pth_SQ.q_lock), tid);
    for (t = pth_pqueue_head(&pth_SQ); t != NULL;
         t = pth_pqueue_walk(&pth_SQ, t, PTH_WALK_NEXT)) {
        sigandset(&tmp, &tmp, &t->mctx.sigs);
    }
    _pth_release_lock(&(pth_SQ.q_lock), tid);

    slot = 0;
    while (pth_native_list[slot].is_used) {
          if ((t = pth_native_list[slot].current) && (t != pth_native_list[slot].sched))
                sigandset(&tmp, &tmp, &t->mctx.sigs);
          slot++;
    }
 
    /* store all thread's sigmask */
    _pth_acquire_lock(&pth_sig_lock, tid);
    memcpy((void *)&pth_sigblock, &tmp, sizeof(sigset_t));
    _pth_release_lock(&pth_sig_lock, tid);
    return;
}

/* initialize the scheduler ingredients */
intern void pth_scheduler_init(void)
{
    pth_descr_t descr = NULL;

    /* No natives yet... */
    pth_number_of_natives = 0;

    /* initialize pth_native_lock */
    pth_lock_init(pth_native_lock);

    /* initialize pth_key_lock */
    pth_lock_init(pth_key_lock);

    /* initialize pth_scheduler_lock */
    pth_lock_init(pth_scheduler_lock);

    /* initialize pth_sig_lock */
    pth_lock_init(pth_sig_lock);

    pth_threads_queue.th_next = &pth_threads_queue;
    pth_threads_queue.th_prev = &pth_threads_queue;
    pth_threads_count = 0;

    if ((descr = pth_alloc_native(FALSE, FALSE)) == NULL) {
        fprintf(stderr, "**Pth** INIT: Cannot allocate initial native thread descriptor: %s\n",
                strerror(errno));
        abort();
    }

    /* fill in the first native slot... */
    descr->pid		= getpid();
    descr->tid		= gettid();
    descr->stacksize	= 0;
    descr->true_stack	= NULL;
    descr->stack	= NULL;
    descr->sched_index	= 0;
    descr->is_bound     = 1;
    descr->is_used      = TRUE;
    descr->is_running   = TRUE;
    
    /* create the internal signal pipe */
    if (pipe(descr->sigpipe) == -1) {
        fprintf(stderr, "**Pth** INIT: Cannot create internal pipe: %s\n",
                strerror(errno));
        abort();
    }
    pth_fdmode(descr->sigpipe[0], PTH_FDMODE_NONBLOCK);
    pth_fdmode(descr->sigpipe[1], PTH_FDMODE_NONBLOCK);

    /* initalize the thread queues */
    pth_pqueue_init(&pth_NQ);
    pth_pqueue_init(&pth_RQ);
    pth_pqueue_init(&pth_WQ);
    pth_pqueue_init(&pth_SQ);
    pth_pqueue_init(&pth_DQ);

    /* initialize load support */
    pth_loadval = 1.0;
    pth_time_set(&pth_loadticknext, PTH_TIME_NOW);

    sigemptyset(&pth_sigblock);
    return;
}

/* drop all threads (except for the currently active one) */
intern void pth_scheduler_drop(void)
{
    pth_t t;
    int slot = 0;
    pth_descr_t descr = pth_get_native_descr();

    /* Lock all the queues... */
    pth_lock_all();

    /* clear the new queue */
    while ((t = pth_pqueue_delmax(&pth_NQ)) != NULL)
        pth_tcb_free(t);
    pth_pqueue_init(&pth_NQ);

    /* clear the ready queue */
    while ((t = pth_pqueue_delmax(&pth_RQ)) != NULL)
        pth_tcb_free(t);
    pth_pqueue_init(&pth_RQ);

    while ((t = pth_pqueue_delmax(&pth_WQ)) != NULL)
        pth_tcb_free(t);
    pth_pqueue_init(&pth_WQ);

    /* clear the waiting queue */
    _pth_acquire_lock(&pth_native_lock, descr->tid);
    while (pth_native_list[slot].is_used) {
	pth_descr_t ds = &pth_native_list[slot++];
	_pth_acquire_lock(&(ds->wait_queue.q_lock), descr->tid);
	while ((t = pth_pqueue_delmax(&(ds->wait_queue))) != NULL)
	    pth_tcb_free(t);
	pth_pqueue_init(&(ds->wait_queue));
	_pth_release_lock(&(ds->wait_queue.q_lock), descr->tid);
    }
    _pth_release_lock(&pth_native_lock, descr->tid);

    /* clear the suspend queue */
    while ((t = pth_pqueue_delmax(&pth_SQ)) != NULL)
        pth_tcb_free(t);
    pth_pqueue_init(&pth_SQ);

    /* clear the dead queue */
    while ((t = pth_pqueue_delmax(&pth_DQ)) != NULL)
        pth_tcb_free(t);
    pth_pqueue_init(&pth_DQ);

    /* Release all the queues... */
    pth_release_all();

    return;
}

intern int pth_WQ_count(void)
{
    int slot = 0;
    int count = 0;

    while (pth_native_list[slot].is_used) {
	pth_descr_t ds = &pth_native_list[slot++];
	count += pth_pqueue_elements(&(ds->wait_queue));
    }
    count += pth_pqueue_elements(&pth_WQ);
    return count;
}

intern int pth_WQ_contains(pth_t t)
{
    int rc = FALSE;
    int slot = 0;

    while (pth_native_list[slot].is_used) {
	pth_descr_t ds = &pth_native_list[slot++];
	if (pth_pqueue_contains(&(ds->wait_queue), t)) {
	    rc = TRUE;
	    break;
	}
    }
    return rc;
}

intern int pth_WQ_delete(pth_t t)
{
    int rc = FALSE;
    int slot = 0;
    pid_t tid = gettid();

    while (pth_native_list[slot].is_used) {
	pth_descr_t ds = &pth_native_list[slot++];
	if (pth_pqueue_contains(&(ds->wait_queue), t)) {
	    rc = TRUE;
	    _pth_acquire_lock(&(ds->wait_queue.q_lock), tid);
	    pth_pqueue_delete(&(ds->wait_queue), t);
	    _pth_release_lock(&(ds->wait_queue.q_lock), tid);
	    break;
	}
    }
    return rc;
}

/*
 * Update the average scheduler load.
 *
 * This is called on every context switch, but we have to adjust the
 * average load value every second, only. When we're called more than
 * once per second we handle this by just calculating anything once
 * and then do NOPs until the next ticks is over. When the scheduler
 * waited for more than once second (or a thread CPU burst lasted for
 * more than once second) we simulate the missing calculations. That's
 * no problem because we can assume that the number of ready threads
 * then wasn't changed dramatically (or more context switched would have
 * been occurred and we would have been given more chances to operate).
 * The actual average load is calculated through an exponential average
 * formula.
 */
#define pth_scheduler_load(now) \
    if (pth_time_cmp((now), &pth_loadticknext) >= 0) { \
        pth_time_t ttmp; \
        int numready; \
        numready = pth_pqueue_elements(&pth_RQ); \
        pth_time_set(&ttmp, (now)); \
        do { \
            pth_loadval = (numready*0.25) + (pth_loadval*0.75); \
            pth_time_sub(&ttmp, &pth_loadtickgap); \
        } while (pth_time_cmp(&ttmp, &pth_loadticknext) >= 0); \
        pth_time_set(&pth_loadticknext, (now)); \
        pth_time_add(&pth_loadticknext, &pth_loadtickgap); \
    }

/* the heart of this library: the thread scheduler */
intern void *pth_scheduler(void *dummy)
{
    sigset_t sigs;
    pth_time_t running;
    pth_time_t snapshot;
    struct sigaction sa;
    sigset_t ss;
    int sig;
    pth_t t;
    pth_descr_t descr = NULL;
    volatile pth_t current;
    volatile pth_t this_sched = NULL;

    /*
     * bootstrapping
     */
    /* find the pth_t for this scheduler... */
    if ((descr = pth_get_native_descr()) == NULL) {
	fprintf(stderr,"pth_scheduler: unable to find scheduler for pid %i.  Aborting...\n", 
		(unsigned int)gettid());
	abort();
    }
    this_sched = descr->sched;
    pth_debug1("pth_scheduler: bootstrapping");

    /* mark this thread as a special scheduler thread */
    this_sched->state = PTH_STATE_SCHEDULER;

    /* block all signals in the scheduler thread */
    sigfillset(&sigs);
    pth_sc(sigprocmask)(SIG_SETMASK, &sigs, NULL);

    /* initialize the snapshot time for bootstrapping the loop */
    pth_time_set(&snapshot, PTH_TIME_NOW);

    /*
     * endless scheduler loop
     */
    for (;;) {
	pth_debug1("pth_scheduler: running...");

        /*
         * Move threads from new queue to ready queue and give
         * them maximum priority so they start immediately
         */
	_pth_acquire_lock(&(pth_NQ.q_lock), descr->tid);
	while ((t = pth_pqueue_tail(&pth_NQ)) != NULL) {
	    pth_pqueue_delete(&pth_NQ, t);
	    t->state = PTH_STATE_READY;
	    _pth_acquire_lock(&(pth_RQ.q_lock), descr->tid);
	    pth_pqueue_insert(&pth_RQ, pth_pqueue_favorite_prio(&pth_RQ), t);
	    _pth_release_lock(&(pth_RQ.q_lock), descr->tid);
	    pth_debug3("pth_scheduler: new thread(0x%lx) \"%s\" moved to top of ready queue", 
	               t, t->name);
	}
	_pth_release_lock(&(pth_NQ.q_lock), descr->tid);

        /*
         * Update average scheduler load
         */
        pth_scheduler_load(&snapshot);

	/*ibm begin*/
	while(TRUE) {
	    /*
	     * Find next thread in ready queue
	     */
	    _pth_acquire_lock(&(pth_RQ.q_lock), descr->tid);
	    current = pth_pqueue_delmax(&pth_RQ);
	    if (current == NULL) {
		pth_debug1("pth_scheduler: No threads ready to run on this native thread, sleeping...\n");
		_pth_release_lock(&(pth_RQ.q_lock), descr->tid);
		pth_native_yield();
		pth_debug1("pth_scheduler: Awake again, looking for work...");
		break;
	    }

	    /*
	     * See if the thread is unbound...
	     * Break out and schedule if so...
	     */
	    if (current->boundnative == 0) {
		_pth_release_lock(&(pth_RQ.q_lock), descr->tid);
		break;
	    }
	    /*
	     * See if the thread is bound to a different native thread...
	     * Break out and schedule if not...
	     */
	    if (current->boundnative == this_sched->lastrannative) {
		_pth_release_lock(&(pth_RQ.q_lock), descr->tid);
		break;
	    }

	    /* 
	     * The thread is bound to a different native thread...
	     * We, therefore, need to put it back in the ready queue but
	     * we'll give it favored status...
	     */
            pth_pqueue_insert(&pth_RQ, pth_pqueue_favorite_prio(&pth_RQ), current);
	    _pth_release_lock(&(pth_RQ.q_lock), descr->tid);

	    /* if current boundnative thread is sleeping, wakeup */
	    if (current->boundnative) {
		char c=(int)1;
		pth_descr_t ds = pth_find_native_by_tid(current->boundnative);
 
		if (ds && ds->is_bound == FALSE)
		    pth_sc(write)(ds->sigpipe[1], &c, sizeof(char));
		current = NULL;
		break;
	    }
	}

	if (current == NULL) 
	    goto event_wait;


	/*ibm end*/
        pth_debug5("pth_scheduler: thread(0x%lx) \"%s\" selected (prio=%d, qprio=%d)",
                   current, current->name, current->prio, current->q_prio);

        /*
         * Raise additionally thread-specific signals
         * (they are delivered when we switch the context)
         *
         * Situation is ('#' = signal pending):
         *     process pending (descr->sigpending):      ----####
         *     thread pending (pth_current->sigpending): --##--##
         * Result has to be:
         *     process new pending:                      --######
         */
	_pth_acquire_lock(&(current->lock), descr->tid);
        if (current->sigpendcnt > 0) {
	    int cnt = 1;
            sigpending(&descr->sigpending);
            for (sig = 1; sig < PTH_NSIG; sig++) {
                if (sigismember(&current->sigpending, sig)) {
                    if (!sigismember(&descr->sigpending, sig)) {
			if (sigismember(&current->sigactionmask, sig))
                        	tkill(descr->tid, sig); /* send to a thread */
			else
                        	kill(descr->pid, sig);	/* send to process */
		    }
		    if (cnt++ == current->sigpendcnt)
			break;
		}
	    }
        }
	_pth_release_lock(&(current->lock), descr->tid);

        /*
         * Set running start time for new thread
         * and perform a context switch to it
         */
        pth_debug3("pth_scheduler: switching to thread 0x%lx (\"%s\")",
                   (unsigned long)current, current->name);

        /* update thread times */
        pth_time_set(&current->lastran, PTH_TIME_NOW);

        /* update scheduler times */
        pth_time_set(&running, &current->lastran);
        pth_time_sub(&running, &snapshot);
        pth_time_add(&this_sched->running, &running);

	/* update the native thread we're running on this time...     ibm*/
	current->lastrannative = descr->tid;

	/* set the "current" thread as current */
	descr->current = current;

    	_pth_release_lock(&descr->sched_lock, descr->tid);

        /* ** ENTERING THREAD ** - by switching the machine context */
        pth_mctx_switch(&this_sched->mctx, &current->mctx);

    	_pth_acquire_lock(&descr->sched_lock, descr->tid);
	/* set the scheduler thread as current */
	descr->current = this_sched;

        /* update scheduler times */
        pth_time_set(&snapshot, PTH_TIME_NOW);
        pth_debug3("pth_scheduler: cameback from thread 0x%lx (\"%s\")",
                   (unsigned long)current, current->name);

        /*
         * Calculate and update the time the previous thread was running
         */
        pth_time_set(&running, &snapshot);
        pth_time_sub(&running, &current->lastran);
        pth_time_add(&current->running, &running);
        pth_debug4("pth_scheduler: thread (0x%lx) \"%s\" ran %.6f",
                   current, current->name, pth_time_t2d(&running));


        /*
         * Remove still pending thread-specific signals
         * (they are re-delivered next time)
         *
         * Situation is ('#' = signal pending):
         *     thread old pending (pth_current->sigpending): --##--##
         *     process old pending (descr->sigpending):      ----####
         *     process still pending (sigstillpending):      ---#-#-#
         * Result has to be:
         *     process new pending:                          -----#-#
         *     thread new pending (pth_current->sigpending): ---#---#
         */
	_pth_acquire_lock(&(current->lock), descr->tid);
        if (current->sigpendcnt > 0) {
            sigset_t sigstillpending;
            sigpending(&sigstillpending);
            for (sig = 1; sig < PTH_NSIG; sig++) {
                if ((current->sigthread != current) && sigismember(&current->sigpending, sig)) {
		    /* if the current thread is signalled by another thread... */
                    if (!sigismember(&sigstillpending, sig)) {
                        /* thread (and perhaps also process) signal delivered */
                        sigdelset(&current->sigpending, sig);
			if (!current->sigpendcnt--)
		    	     break;
                    }
                    else if (!sigismember(&descr->sigpending, sig)) {
                        /* thread signal not delivered */
                        pth_util_sigdelete(sig);
                    }
		}
            }
        }
	_pth_release_lock(&(current->lock), descr->tid);

        /*
         * Check for stack overflow
         */
        if (current->stackguard != NULL) {
            if (*current->stackguard != 0xDEAD) {
                pth_debug3("pth_scheduler: stack overflow detected for thread 0x%lx (\"%s\")",
                           (unsigned long)current, current->name);
                /*
                 * if the application doesn't catch SIGSEGVs, we terminate
                 * manually with a SIGSEGV now, but output a reasonable message.
                 */
                if (sigaction(SIGSEGV, NULL, &sa) == 0) {
                    if (sa.sa_handler == SIG_DFL) {
                        fprintf(stderr, "**NGPT** STACK OVERFLOW: tid = %d, thread pid_t=0x%lx, name=\"%s\"\n",
                                (int) current_tid(), (unsigned long)current, current->name);
			/* We kill the main thread, passign the segv... */
                        kill(pth_primordial_thread()->tid, SIGSEGV);
                        sigfillset(&ss);
                        sigdelset(&ss, SIGSEGV);
                        sigsuspend(&ss);
                        abort();
                    }
                }
                /*
                 * else we terminate the thread only and send us a SIGSEGV
                 * which allows the application to handle the situation...
                 */
                current->join_arg = (void *)0xDEAD;
                current->state = PTH_STATE_EXIT;
                tkill(descr->tid, SIGSEGV);
            }
        }

        /*
         * When previous thread is now marked as dead, kick it out
         */
	if (current->state == PTH_STATE_DEAD)
	    current = NULL;
        if (current != NULL && current->state == PTH_STATE_EXIT) {
            pth_debug3("pth_scheduler: marking thread (0x%lx) \"%s\" as exiting", current, current->name);
            if (!current->joinable) {
		pth_debug2("pth_scheduler: thread 0x%lx not joinable, reaping...", current);
                pth_tcb_free(current); 
	    } else {
		pth_debug2("pth_scheduler: thread 0x%lx joinable, moving to DEAD queue...", current);
		_pth_acquire_lock(&(pth_DQ.q_lock), descr->tid);
                pth_pqueue_Qinsert(&pth_DQ, PTH_PRIO_STD, current);
	        _pth_release_lock(&(pth_DQ.q_lock), descr->tid);
		current->state = PTH_STATE_DEAD;
	    }
            current = NULL;
        }

        /*
         * When thread wants to wait for an event
         * move it to waiting queue now
         */
        if (current != NULL && current->state == PTH_STATE_WAITING) {
            pth_debug3("pth_scheduler: moving thread (0x%lx) \"%s\" to waiting queue", current, current->name);
	    if ((current == pth_main) && (descr != &pth_native_list[0])) {
		char c = (int)1;
		pth_descr_t ds = &pth_native_list[0];
		_pth_acquire_lock(&(ds->wait_queue.q_lock), descr->tid);
		pth_pqueue_Qinsert(&ds->wait_queue, 0, current);
		_pth_release_lock(&(ds->wait_queue.q_lock), descr->tid);
		pth_sc(write)(ds->sigpipe[1], &c, sizeof(char));
	    } else {
		_pth_acquire_lock(&(descr->wait_queue.q_lock), descr->tid);
		pth_pqueue_Qinsert(&(descr->wait_queue), 0, current);
		_pth_release_lock(&(descr->wait_queue.q_lock), descr->tid);
	    }
            current = NULL;
        }

        /*
         * migrate old treads in ready queue into higher
         * priorities to avoid starvation and insert last running
         * thread back into this queue, too.
         */
	_pth_acquire_lock(&(pth_RQ.q_lock), descr->tid);
        pth_pqueue_increase(&pth_RQ);
        if (current != NULL)
            pth_pqueue_Qinsert(&pth_RQ, current->prio, current);
	_pth_release_lock(&(pth_RQ.q_lock), descr->tid);

        /*
         * Manage the events in the waiting queue, i.e. decide whether their
         * events occurred and move them to the ready queue. But wait only if
         * we have already no new or ready threads.
         */
    event_wait:
  	pth_time_set(&snapshot, PTH_TIME_NOW);
        if (   pth_pqueue_elements(&pth_RQ) == 0
            && pth_pqueue_elements(&pth_NQ) == 0 )
            pth_sched_eventmanager(&snapshot, descr, FALSE /* wait */);
        else
	    pth_sched_eventmanager(&snapshot, descr, TRUE  /* poll */);
    }

    /* NOTREACHED */
    return NULL;
}

/*
 * Look whether some events already occurred and move
 * corresponding threads from waiting queue back to ready queue.
 */
intern void pth_sched_eventmanager(pth_time_t *now, pth_descr_t descr, int dopoll)
{
    pth_event_t nexttimer_ev;
    pth_time_t nexttimer_value;
    pth_event_t evh;
    pth_event_t ev;
    pth_t t = NULL;
    pth_t tlast;
    int this_occurred;
    int any_occurred;
    fd_set rfds;
    fd_set wfds;
    fd_set efds;
    struct timeval delay;
    struct timeval *pdelay;
    sigset_t oss;
    struct sigaction sa;
    struct sigaction osa[1+PTH_NSIG];
    char minibuf[128];
    int loop_repeat;
    int fdmax;
    int n;
    int rc;
    int sig;
    int slot = 0;
    int	any_event_occurred;
    sigset_t sigs_we_block;
    int mutex_event;

    pth_debug2("pth_sched_eventmanager: enter in %s mode",
               dopoll ? "polling" : "waiting");

    /* entry point for internal looping in event handling */
    loop_entry:
    loop_repeat = FALSE;

    /* initialize fd sets */
    FD_ZERO(&rfds);
    FD_ZERO(&wfds);
    FD_ZERO(&efds);
    fdmax = -1;

    /* initialize signal status */
    sigpending(&descr->sigpending);
    sigfillset(&descr->sigblock);
    sigemptyset(&descr->sigcatch);
    sigemptyset(&descr->sigraised);

    /* initialize next timer */
    pth_time_set(&nexttimer_value, PTH_TIME_ZERO);
    descr->nexttimer_thread = NULL;
    nexttimer_ev = NULL;

    if (descr == &pth_native_list[0]) {
    	_pth_acquire_lock(&pth_sig_lock, descr->tid);
    	memcpy((void *)&sigs_we_block, &pth_sigblock, sizeof(sigset_t)); 
    	_pth_release_lock(&pth_sig_lock, descr->tid);
    }

    any_event_occurred = FALSE;
    mutex_event = FALSE;

    /* for all threads in the native's waiting queue... */
    while (pth_native_list[slot].is_used) {
	pth_descr_t ds = &pth_native_list[slot++];
	/* if dopoll, examine own waiting queue's events... */
	if (dopoll)
	    ds = descr;
	_pth_acquire_lock(&(ds->wait_queue.q_lock), descr->tid);
	t = pth_pqueue_head(&(ds->wait_queue));
	while (t != NULL) {

	    any_occurred = FALSE;

	    /* ... and all their events... */
	    if (t->events == NULL) {
		t = pth_pqueue_walk(&(ds->wait_queue), t, PTH_WALK_NEXT);
		continue;
	    }

	    /* ...check whether events occurred */
	    ev = evh = t->events;
	    do {
		/* cancellation support... */
		if (t->cancelreq && (ev->ev_type != PTH_EVENT_FUNC)) {
		    any_occurred = TRUE;
		    break;
		}
		if (!ev->ev_occurred) {
		    this_occurred = FALSE;

		    /* Filedescriptor I/O */
		    if (ev->ev_type == PTH_EVENT_FD) {
			/* filedescriptors are checked later all at once.
			   Here we only assemble them in the fd sets */
			if (ev->ev_goal & PTH_UNTIL_FD_READABLE)
			    FD_SET(ev->ev_args.FD.fd, &rfds);
			if (ev->ev_goal & PTH_UNTIL_FD_WRITEABLE)
			    FD_SET(ev->ev_args.FD.fd, &wfds);
			if (ev->ev_goal & PTH_UNTIL_FD_EXCEPTION)
			    FD_SET(ev->ev_args.FD.fd, &efds);
			if (fdmax < ev->ev_args.FD.fd)
			    fdmax = ev->ev_args.FD.fd;
		    }
		    /* Filedescriptor Set Select I/O */
		    else if (ev->ev_type == PTH_EVENT_SELECT) {
			/* filedescriptors are checked later all at once.
			   Here we only merge the fd sets. */
			pth_util_fds_merge(ev->ev_args.SELECT.nfd,
					   ev->ev_args.SELECT.rfds, &rfds,
					   ev->ev_args.SELECT.wfds, &wfds,
					   ev->ev_args.SELECT.efds, &efds);
			if (fdmax < ev->ev_args.SELECT.nfd-1)
			    fdmax = ev->ev_args.SELECT.nfd-1;
		    }
		    /* Signal Set */
		    else if (ev->ev_type == PTH_EVENT_SIGS) {
			for (sig = 1; sig < PTH_NSIG; sig++) {
			    if (sigismember(ev->ev_args.SIGS.sigs, sig)) {
				/* thread signal handling */
				if (sigismember(&t->sigpending, sig)) {
				    if (ev->ev_flags == TRUE) {
				        *(ev->ev_args.SIGS.sig) = sig;
				    	this_occurred = TRUE;
				        sigdelset(&t->sigpending, sig);
					if (!t->sigpendcnt--)
					     break;
				    } else {	/* for sleep() */
					/* if sig is blocked, don't interrupt sleep */
					if (sigismember(&t->mctx.sigs, sig))
						continue;
				        *(ev->ev_args.SIGS.sig) = 0;
				    	this_occurred = TRUE;
					break;
				    }
				}
				/* process signal handling */
				if (sigismember(&descr->sigpending, sig)) {
				    if (ev->ev_args.SIGS.sig != NULL)
					*(ev->ev_args.SIGS.sig) = sig;
				    pth_util_sigdelete(sig);
				    sigdelset(&descr->sigpending, sig);
				    this_occurred = TRUE;
				}
				else {
				    if (descr == &pth_native_list[0]) {
					sigdelset(&sigs_we_block, sig);
					sigaddset(&descr->sigcatch, sig);
				    }
				}
			    }
			}
		    }
		    /* Timer */
		    else if (ev->ev_type == PTH_EVENT_TIME) {
			if (pth_time_cmp(&(ev->ev_args.TIME.tv), now) < 0)
			    this_occurred = TRUE;
			else {
			    /* remember the timer which will be elapsed next */
			    if (ev->ev_flags != TRUE) {
				if ((descr->nexttimer_thread == NULL && nexttimer_ev == NULL) ||
				    pth_time_cmp(&(ev->ev_args.TIME.tv), &nexttimer_value) < 0) {
				    if (nexttimer_ev)
				    	nexttimer_ev->ev_flags = FALSE;
				    descr->nexttimer_thread = t;
				    nexttimer_ev = ev;
				    nexttimer_ev->ev_flags = TRUE;
				    pth_time_set(&nexttimer_value, &(ev->ev_args.TIME.tv));
				}
			    }
			}
		    }
		    /* Message Port Arrivals */
		    else if (ev->ev_type == PTH_EVENT_MSG) {
			if (pth_ring_elements(&(ev->ev_args.MSG.mp->mp_queue)) > 0)
			    this_occurred = TRUE;
		    }
		    /* Mutex Release */
		    else if (ev->ev_type == PTH_EVENT_MUTEX) {
			if (!(ev->ev_args.MUTEX.mutex->mx_state & PTH_MUTEX_LOCKED)) {
			    this_occurred = TRUE;
			    mutex_event = TRUE;
			}
		    }
		    /* Condition Variable Signal */
		    else if (ev->ev_type == PTH_EVENT_COND) {
			if (ev->ev_args.COND.cond->cn_state & PTH_COND_SIGNALED) {
			    if (ev->ev_args.COND.cond->cn_state & PTH_COND_BROADCAST)
				this_occurred = TRUE;
			    else {
				if (!(ev->ev_args.COND.cond->cn_state & PTH_COND_HANDLED)) {
					ev->ev_args.COND.cond->cn_state |= PTH_COND_HANDLED;
					this_occurred = TRUE;
					mutex_event = TRUE;
				}
			    }
			}
		    }
		    /* Thread Termination */
		    else if (ev->ev_type == PTH_EVENT_TID) {
			if (   (   ev->ev_args.TID.tid == NULL
				&& pth_pqueue_elements(&pth_DQ) > 0)
			    || (   ev->ev_args.TID.tid != NULL
				&& ev->ev_args.TID.tid->state == ev->ev_goal)
			    || (t == pth_main && pth_active_threads == 1
				&& !pth_thread_exists(ev->ev_args.TID.tid)))
			    this_occurred = TRUE;
		    }
		    /* Custom Event Function */
		    else if (ev->ev_type == PTH_EVENT_FUNC) {
			if (ev->ev_args.FUNC.func(ev->ev_args.FUNC.arg)) 
			    this_occurred = TRUE;
			else if (ev->ev_flags != TRUE) {
			    pth_time_t tv;
			    pth_time_set(&tv, now);
			    pth_time_add(&tv, &(ev->ev_args.FUNC.tv));
			    if ((descr->nexttimer_thread == NULL && nexttimer_ev == NULL) ||
				pth_time_cmp(&tv, &nexttimer_value) < 0) {
				if (nexttimer_ev)
				    nexttimer_ev->ev_flags = FALSE;
				descr->nexttimer_thread = t;
				nexttimer_ev = ev;
				nexttimer_ev->ev_flags = TRUE;
				pth_time_set(&nexttimer_value, &tv);
			    }
			}
		    }

		    /* tag event if it has occurred */
		    if (this_occurred) {
			pth_debug2("pth_sched_eventmanager: [non-I/O] event occurred for thread \"%s\"", t->name);
			ev->ev_occurred = TRUE;
			any_occurred = TRUE;
		    }
		}
	    } while ((ev = ev->ev_next) != evh);
	    tlast = t;
	    t = pth_pqueue_walk(&(ds->wait_queue), t, PTH_WALK_NEXT);
	    if (any_occurred) {
		pth_pqueue_delete(&(ds->wait_queue), tlast);
		tlast->state = PTH_STATE_READY;
		_pth_acquire_lock(&(pth_RQ.q_lock), descr->tid);
		pth_pqueue_insert(&pth_RQ, tlast->prio+1, tlast);
		_pth_release_lock(&(pth_RQ.q_lock), descr->tid);
		any_event_occurred = TRUE;
	    }
	}
	_pth_release_lock(&((ds->wait_queue).q_lock), descr->tid);
	if (dopoll || mutex_event)
	    break;
    }

    if (any_event_occurred)
        dopoll = TRUE;

    /* now decide how to poll for fd I/O and timers */
    if (dopoll) {
        /* do a polling with immediate timeout,
           i.e. check the fd sets only without blocking */
        pth_time_set(&delay, PTH_TIME_ZERO);
        pdelay = &delay;
	if (descr->nexttimer_thread != NULL && nexttimer_ev != NULL)
	    nexttimer_ev->ev_flags = FALSE;
    }
    else if (nexttimer_ev != NULL) {
        /* do a polling with a timeout set to the next timer,
           i.e. wait for the fd sets or the next timer */
        pth_time_set(&delay, &nexttimer_value);
        pth_time_sub(&delay, now);
        pdelay = &delay;
    }
    else {
        /* do a polling without a timeout,
           i.e. wait for the fd sets only with blocking */
        pdelay = NULL;
    }

    /* clear pipe and let select() wait for the read-part of the pipe */
    while (pth_sc(read)(descr->sigpipe[0], minibuf, sizeof(minibuf)) > 0) ;
    FD_SET(descr->sigpipe[0], &rfds);

    /* If dopoll is set, handle the thread from ready queue... */
    if (dopoll)
	return;
    
    if (fdmax < descr->sigpipe[0])
        fdmax = descr->sigpipe[0];

    /* replace signal actions for signals we've to catch for events */
    if (descr == &pth_native_list[0]) {
    	for (sig = 1; sig < PTH_NSIG; sig++) {
	    if (sigismember(&descr->sigcatch, sig)) {
            	sa.sa_handler = pth_sched_eventmanager_sighandler;
            	sigfillset(&sa.sa_mask);
            	sa.sa_flags = 0;
            	sigaction(sig, &sa, &osa[sig]);
	    }
    	}
    }

    /* allow some signals to be delivered: Either to our
       catching handler or directly to the configured
       handler for signals not catched by events */
    if (descr == &pth_native_list[0])
    	pth_sc(sigprocmask)(SIG_SETMASK, &sigs_we_block, &oss);

    /* now do the polling for filedescriptor I/O and timers
       WHEN THE SCHEDULER SLEEPS AT ALL, THEN HERE!! */
    rc = -1;
    descr->is_bound = 0;
    if (!(dopoll && fdmax == -1)) {
	_pth_release_lock(&(descr->sched_lock), descr->tid);

        while ((rc = pth_sc(select)(fdmax+1, &rfds, &wfds, &efds, pdelay)) < 0
               && errno == EINTR) ;

	_pth_acquire_lock(&(descr->sched_lock), descr->tid);
	if (descr->nexttimer_thread == NULL)
	    nexttimer_ev = NULL;
    }
    descr->is_bound = 1;

    /* restore signal mask and actions and handle signals */
    if (descr == &pth_native_list[0]) {
    	pth_sc(sigprocmask)(SIG_SETMASK, &oss, NULL);
    	for (sig = 1; sig < PTH_NSIG; sig++) {
	    if (sigismember(&descr->sigcatch, sig))
            	sigaction(sig, &osa[sig], NULL);
	}
    }

    /* if the timer elapsed, handle it */
    if (!dopoll && nexttimer_ev != NULL) {
	if (rc == 0) {
	    pth_debug2("pth_sched_eventmanger: nexttimer_ev = 0x%lx", nexttimer_ev);
	    if (nexttimer_ev->ev_type == PTH_EVENT_FUNC) {
		/* it was an implicit timer event for a function event,
		   so repeat the event handling for rechecking the function */
		loop_repeat = TRUE;
	    }
	    else {
		/* it was an explicit timer event, standing for its own */
		pth_debug3("pth_sched_eventmanager:[timeout] event occurred for thread 0x%lx \"%s\"",
			   descr->nexttimer_thread, descr->nexttimer_thread->name);
		nexttimer_ev->ev_flags = FALSE;
		nexttimer_ev->ev_occurred = TRUE;

	    }
	} else {
            /* timer has not elapsed, clear flag or it might get lost - aj */
            pth_debug3("pth_sched_eventmanager: resetting non elapsed timer for thread 0x%lx \"%s\"", descr->nexttimer_thread, descr->nexttimer_thread->name);
	    if (descr->nexttimer_thread != NULL && nexttimer_ev != NULL)
            	nexttimer_ev->ev_flags = FALSE;
	}
    }

    /* if the internal signal pipe was used, adjust the select() results */
    if (!dopoll && rc > 0 && FD_ISSET(descr->sigpipe[0], &rfds)) {
        FD_CLR(descr->sigpipe[0], &rfds);
        rc--;
    }

    /* if an error occurred, avoid confusion in the cleanup loop */
    if (rc <= 0) {
        FD_ZERO(&rfds);
        FD_ZERO(&wfds);
        FD_ZERO(&efds);
    }

    /* now comes the final cleanup loop where we've to
       do two jobs: first we've to the late handling of the fd I/O events and
       additionally if a thread has one occurred event, we move it from the
       waiting queue to the ready queue */

    /* for all threads in the waiting queue... */
    slot = 0;
    while (pth_native_list[slot].is_used) {
	pth_descr_t ds = &pth_native_list[slot++];
	_pth_acquire_lock(&(ds->wait_queue.q_lock), descr->tid);
	t = pth_pqueue_head(&ds->wait_queue);
    
	while (t != NULL) {

	    /* do the late handling of the fd I/O and signal
	       events in the waiting event ring */
	    any_occurred = FALSE;
	    if (t->events != NULL) {
		ev = evh = t->events;
		do {
		    /*
		     * Late handling for still not occured events
		     */
		    if (!ev->ev_occurred) {
			/* Filedescriptor I/O */
			if (ev->ev_type == PTH_EVENT_FD) {
			    if (   (   ev->ev_goal & PTH_UNTIL_FD_READABLE
				    && FD_ISSET(ev->ev_args.FD.fd, &rfds))
				|| (   ev->ev_goal & PTH_UNTIL_FD_WRITEABLE
				    && FD_ISSET(ev->ev_args.FD.fd, &wfds))
				|| (   ev->ev_goal & PTH_UNTIL_FD_EXCEPTION
				    && FD_ISSET(ev->ev_args.FD.fd, &efds)) ) {
				pth_debug3("pth_sched_eventmanager: "
					   "[I/O] event occurred for thread 0x%lx \"%s\"", t, t->name);
				ev->ev_occurred = TRUE;
			    }
			}
			/* Filedescriptor Set I/O */
			else if (ev->ev_type == PTH_EVENT_SELECT) {
			    if (pth_util_fds_test(ev->ev_args.SELECT.nfd,
						  ev->ev_args.SELECT.rfds, &rfds,
						  ev->ev_args.SELECT.wfds, &wfds,
						  ev->ev_args.SELECT.efds, &efds)) {
				n = pth_util_fds_select(ev->ev_args.SELECT.nfd,
							ev->ev_args.SELECT.rfds, &rfds,
							ev->ev_args.SELECT.wfds, &wfds,
							ev->ev_args.SELECT.efds, &efds);
				if (ev->ev_args.SELECT.n != NULL)
				    *(ev->ev_args.SELECT.n) = n;
				ev->ev_occurred = TRUE;
				pth_debug3("pth_sched_eventmanager: "
					   "[I/O] event occurred for thread 0x%lx \"%s\"", t, t->name);
			    }
			}
			/* Signal Set */
			else if (ev->ev_type == PTH_EVENT_SIGS) {
			    for (sig = 1; sig < PTH_NSIG; sig++) {
				if (sigismember(ev->ev_args.SIGS.sigs, sig)) {
				    if (sigismember(&descr->sigraised, sig)) {
					if (ev->ev_args.SIGS.sig != NULL)
					    *(ev->ev_args.SIGS.sig) = sig;
					pth_debug3("pth_sched_eventmanager: "
						   "[signal] event occurred for thread 0x%lx \"%s\"", t, t->name);
					sigdelset(&descr->sigraised, sig);
					ev->ev_occurred = TRUE;
				    }
				}
			    }
			}
		    }

		    /* local to global mapping */
		    if (ev->ev_occurred)
			any_occurred = TRUE;
		} while ((ev = ev->ev_next) != evh);
	    }

	    /* cancellation support */
	    if (t->cancelreq == TRUE) {
		pth_debug3("pth_sched_eventmanager: cancellation request pending for thread 0x%lx \"%s\"", t, t->name);
		any_occurred = TRUE;
	    }

	    /* Get the next entry... */
	    tlast = t;
	    t = pth_pqueue_walk(&(ds->wait_queue), t, PTH_WALK_NEXT);

	    /*
	     * move last thread to ready queue if any events occurred for it.
	     * we insert it with a slightly increased queue priority to it a
	     * better chance to immediately get scheduled, else the last running
	     * thread might immediately get again the CPU which is usually not
	     * what we want, because we oven use pth_yield() calls to give others
	     * a chance.
	     */
	    if (any_occurred) {
		pth_pqueue_delete(&(ds->wait_queue), tlast);
		tlast->state = PTH_STATE_READY;
		_pth_acquire_lock(&(pth_RQ.q_lock), descr->tid);
		pth_pqueue_insert(&pth_RQ, tlast->prio+1, tlast);
		_pth_release_lock(&(pth_RQ.q_lock), descr->tid);
		pth_debug3("pth_sched_eventmanager: thread (0x%lx) \"%s\" moved from waiting "
			   "to ready queue", tlast, tlast->name);
	    }
	}

	_pth_release_lock(&(ds->wait_queue.q_lock), descr->tid);
    }
    
    /* perhaps we have to internally loop... */
    if (loop_repeat) {
	pth_time_set(now, PTH_TIME_NOW);
        goto loop_entry;
    }

    pth_debug1("pth_sched_eventmanager: leaving");
    return;
}

intern void pth_sched_eventmanager_sighandler(int sig)
{
    char c;
    pth_descr_t descr = NULL;

    pth_debug2("pth_sched_eventmanger_sighandler: caught signal %d", sig);

    /* Get the thread descriptor for the running thread... */
    if ((descr = pth_get_native_descr()) == NULL) {
	fprintf(stderr, "pth_sched_eventmanager_sighandler: no scheduler found !?!?!\n");
	abort();
    }

    /* remember raised signal */
    sigaddset(&descr->sigraised, sig);

    /* write signal to signal pipe in order to awake the select() */
    c = (int)sig;
    pth_sc(write)(pth_native_list[0].sigpipe[1], &c, sizeof(char));
    return;
}

